@@ -7,11 +7,10 @@ from explainaboard.loaders import (
     language_modeling,
     loader,
     loader_registry,
-    named_entity_recognition,
     qa_multiple_choice,
+    sequence_labeling,
     text_classification,
     text_pair_classification,
-    word_segmentation,
 )
 
 get_datalab_loader = loader_registry.get_datalab_loader
@@ -25,9 +24,8 @@ __all__ = [
     'kg_link_tail_prediction',
     'language_modeling',
     'loader',
-    'named_entity_recognition',
+    'sequence_labeling',
     'qa_multiple_choice',
     'text_classification',
     'text_pair_classification',
-    'word_segmentation',
 ]
