@@ -14,11 +14,11 @@ import explainaboard.utils.feature_funcs
 from explainaboard.utils.typing_utils import unwrap
 
 
-@register_processor(TaskType.question_answering_extractive)
+@register_processor(TaskType.qa_extractive)
 class QAExtractiveProcessor(Processor):
     @classmethod
     def task_type(cls) -> TaskType:
-        return TaskType.question_answering_extractive
+        return TaskType.qa_extractive
 
     @classmethod
     def default_features(cls) -> feature.Features:
