@@ -1,23 +1,27 @@
 from __future__ import annotations
 
-from collections.abc import Callable, Iterator
+from collections.abc import Iterator
 import json
 
-from datalabs import aggregating, load_dataset
+from datalabs import aggregating
 
-# TODO(odashi): Add a function to obtain metric class instead of using getattr.
 from explainaboard import feature, TaskType
 from explainaboard.info import SysOutputInfo
 from explainaboard.metric import (
+    Hits,
     HitsConfig,
+    MeanRank,
     MeanRankConfig,
+    MeanReciprocalRank,
     MeanReciprocalRankConfig,
     MetricConfig,
+    MetricStats,
 )
 from explainaboard.processors.processor import Processor
 from explainaboard.processors.processor_registry import register_processor
 from explainaboard.utils import cache_api
-from explainaboard.utils.logging import get_logger, progress
+from explainaboard.utils.logging import progress
+from explainaboard.utils.typing_utils import unwrap
 
 
 @register_processor(TaskType.kg_link_tail_prediction)
@@ -31,8 +35,11 @@ class KGLinkTailPredictionProcessor(Processor):
         return feature.Features(
             {
                 "true_head": feature.Value("string"),
+                "true_head_decipher": feature.Value("string"),
                 "true_link": feature.Value("string"),
                 "true_tail": feature.Value("string"),
+                "true_tail_decipher": feature.Value("string"),
+                "predict": feature.Value("string"),
                 "true_label": feature.Value("string"),
                 "predictions": feature.Sequence(feature=feature.Value("string")),
                 "tail_entity_length": feature.Value(
@@ -113,14 +120,16 @@ class KGLinkTailPredictionProcessor(Processor):
         )
 
     @classmethod
-    def default_metrics(cls, language=None) -> list[MetricConfig]:
+    def default_metrics(
+        cls, source_language=None, target_language=None
+    ) -> list[MetricConfig]:
         return [
-            HitsConfig(name='Hits1', hits_k=1, language=language),
-            HitsConfig(name='Hits2', hits_k=2, language=language),
-            HitsConfig(name='Hits3', hits_k=3, language=language),
-            HitsConfig(name='Hits5', hits_k=5, language=language),
-            MeanReciprocalRankConfig(name='MRR', language=language),
-            MeanRankConfig(name='MR', language=language),
+            HitsConfig(name='Hits1', hits_k=1),
+            HitsConfig(name='Hits2', hits_k=2),
+            HitsConfig(name='Hits3', hits_k=3),
+            HitsConfig(name='Hits5', hits_k=5),
+            MeanReciprocalRankConfig(name='MRR'),
+            MeanRankConfig(name='MR'),
         ]
 
     # TODO: is this the best place to put this?
@@ -140,9 +149,15 @@ class KGLinkTailPredictionProcessor(Processor):
     def __init__(self):
         super().__init__()
         self.entity_type_level_map = None
+        file_path = cache_api.cache_online_file(
+            'http://phontron.com/download/explainaboard/pre_computed/kg/entity_type_level_map.json',  # noqa
+            'pre_computed/kg/entity_type_level_map.json',
+        )
+        with open(file_path, 'r') as file:
+            self.entity_type_level_map = json.load(file)
 
     @aggregating()
-    def _statistics_func(self, samples: Iterator[dict[str, str]]):
+    def _statistics_func(self, samples: Iterator, sys_info: SysOutputInfo):
         """
         `Samples` is a dataset iterator: List[Dict], to know more about it, you can:
         # pip install datalabs
@@ -153,22 +168,45 @@ class KGLinkTailPredictionProcessor(Processor):
         dict_link: dict[str, int] = {}
         dict_tail: dict[str, int] = {}
 
+        entity_dic = {}
+        file_path = cache_api.cache_online_file(
+            'http://phontron.com/download/explainaboard/pre_computed/kg/entity2wikidata.json',  # noqa
+            'pre_computed/kg/entity2wikidata.json',
+        )
+        with open(file_path, 'r') as file:
+            entity_dic = json.loads(file.read())
+
         for sample in progress(samples):
 
-            if sample['tail'] not in dict_tail.keys():
-                dict_tail[sample['tail']] = 1
+            tail = (
+                sample['tail']
+                if sample['tail'] not in entity_dic.keys()
+                else entity_dic[sample['tail']]['label']
+            )
+            if tail not in dict_tail.keys():
+                dict_tail[tail] = 1
             else:
-                dict_tail[sample['tail']] += 1
+                dict_tail[tail] += 1
 
-            if sample['head'] not in dict_head.keys():
-                dict_head[sample['head']] = 1
+            head = (
+                sample['head']
+                if sample['head'] not in entity_dic.keys()
+                else entity_dic[sample['head']]['label']
+            )
+            if head not in dict_head.keys():
+                dict_head[head] = 1
             else:
-                dict_head[sample['head']] += 1
+                dict_head[head] += 1
 
-            if sample['link'] not in dict_link.keys():
-                dict_link[sample['link']] = 1
+            link = (
+                sample['link']
+                if sample['link'] not in entity_dic.keys()
+                else entity_dic[sample['link']]['label']
+            )
+            if link not in dict_link.keys():
+                dict_link[link] = 1
             else:
-                dict_link[sample['link']] += 1
+                dict_link[link] += 1
 
         return {
             "head_fre": dict_head,
@@ -176,46 +214,38 @@ class KGLinkTailPredictionProcessor(Processor):
             "tail_fre": dict_tail,
         }
 
-    def _gen_external_stats(self, sys_info: SysOutputInfo, statistics_func: Callable):
+    def _gen_metric_stats(
+        self, sys_info: SysOutputInfo, sys_output: list[dict]
+    ) -> list[MetricStats]:
+        """Generate sufficient statistics for scoring different metrics.
+        :param sys_info: Information about the system outputs
+        :param sys_output: The system output itself
+        :return: Statistics sufficient for scoring
+        """
 
-        # TODO(gneubig):
-        # this will be reloaded for every dataset, maybe should be fixed for multiple
-        # analysis
-        # TODO(Pengfei): uncomment following code once datalab loader
-        #  of this task  is introduced
-        # if sys_info.dataset_name != "fb15k_237":  # to be generalized
-        #     self.entity_type_level_map = {}
-        # else:
-        self.entity_type_level_map = {}
-        file_path = cache_api.cache_online_file(
-            'http://phontron.com/download/explainaboard/pre_computed/kg/entity_type_level_map.json',  # noqa
-            'pre_computed/kg/entity_type_level_map.json',
-        )
-        with open(file_path, 'r') as file:
-            self.entity_type_level_map = json.load(file)
+        metrics = unwrap(self._get_metrics(sys_info))
+        true_data = [self._get_true_label(x) for x in sys_output]
+        pred_data = [self._get_predicted_label(x) for x in sys_output]
+        rank_data = [
+            self._get_rank_data(x) for x in sys_output
+        ]  # rank of true entity in predictions
 
-        # Calculate statistics of training set
-        self.statistics = None
-        if sys_info.dataset_name is not None:
-            try:
-                dataset = load_dataset(sys_info.dataset_name, "readable")
-                # calculate the statistics (_stat) when _stat is {} or
-                # `reload_stat` is False
-                if len(dataset['train']._stat) == 0 or not sys_info.reload_stat:
-                    new_train = dataset['train'].apply(
-                        self._statistics_func, mode="local"
-                    )
-                    self.statistics = new_train._stat
-                else:
-                    self.statistics = dataset["train"]._stat
-            except FileNotFoundError:
-                get_logger().warning(
-                    """
-The dataset hasn't been supported by DataLab so no training set dependent features will
-be supported by ExplainaBoard. You can add the dataset by:
-https://github.com/ExpressAI/DataLab/blob/main/docs/SDK/add_new_datasets_into_sdk.md
-"""
-                )
+        if any(item is None for item in rank_data):
+            raise ValueError(
+                'Some data points do not have rank information; check system outputs.'
+            )
+
+        metric_stats = []
+        for metric in metrics:
+            if (
+                isinstance(metric, MeanReciprocalRank)
+                or isinstance(metric, MeanRank)
+                or isinstance(metric, Hits)
+            ):
+                metric_stats.append(metric.calc_stats_from_rank(rank_data))
+            else:
+                metric_stats.append(metric.calc_stats_from_data(true_data, pred_data))
+        return metric_stats
 
     # --- Feature functions accessible by ExplainaboardBuilder._get_feature_func()
     def _get_entity_type_level(self, sys_info: SysOutputInfo, existing_features: dict):
@@ -224,7 +254,7 @@ https://github.com/ExpressAI/DataLab/blob/main/docs/SDK/add_new_datasets_into_sd
         # [type_level_0, type_level_1, ... type_level_6]
         # e.g. ["Thing", "Agent", "Person", None, None, None, None]
         tail_entity_type_levels = self.entity_type_level_map.get(
-            existing_features['true_tail_anonymity'], None
+            existing_features['true_tail'], None
         )
         if tail_entity_type_levels is None:
             return "-1"  # entity types not found
@@ -237,37 +267,49 @@ https://github.com/ExpressAI/DataLab/blob/main/docs/SDK/add_new_datasets_into_sd
         return str(most_specific_level)
 
     def _get_tail_entity_length(self, sys_info: SysOutputInfo, existing_features: dict):
-        return len(sys_info.tokenize(existing_features["true_tail"]))
+        return len(
+            unwrap(sys_info.target_tokenizer)(existing_features["true_tail_decipher"])
+        )
 
     def _get_head_entity_length(self, sys_info: SysOutputInfo, existing_features: dict):
-        return len(sys_info.tokenize(existing_features["true_head"]))
+        return len(
+            unwrap(sys_info.source_tokenizer)(existing_features["true_head_decipher"])
+        )
 
-    def _get_tail_fre(self, sys_info: SysOutputInfo, existing_features: dict):
+    def _get_tail_fre(
+        self, sys_info: SysOutputInfo, existing_features: dict, statistics
+    ):
         if (
-            self.statistics is None
-            or existing_features["true_tail"] not in self.statistics['tail_fre'].keys()
+            statistics is None
+            or existing_features["true_tail_decipher"]
+            not in statistics['tail_fre'].keys()
         ):
             return 0
         else:
-            return self.statistics['tail_fre'][existing_features["true_tail"]]
+            return statistics['tail_fre'][existing_features["true_tail_decipher"]]
 
-    def _get_head_fre(self, sys_info: SysOutputInfo, existing_features: dict):
+    def _get_head_fre(
+        self, sys_info: SysOutputInfo, existing_features: dict, statistics
+    ):
         if (
-            self.statistics is None
-            or existing_features["true_head"] not in self.statistics['head_fre'].keys()
+            statistics is None
+            or existing_features["true_head_decipher"]
+            not in statistics['head_fre'].keys()
         ):
             return 0
         else:
-            return self.statistics['head_fre'][existing_features["true_head"]]
+            return statistics['head_fre'][existing_features["true_head_decipher"]]
 
-    def _get_link_fre(self, sys_info: SysOutputInfo, existing_features: dict):
+    def _get_link_fre(
+        self, sys_info: SysOutputInfo, existing_features: dict, statistics
+    ):
         if (
-            self.statistics is None
-            or existing_features["true_link"] not in self.statistics['link_fre'].keys()
+            statistics is None
+            or existing_features["true_link"] not in statistics['link_fre'].keys()
         ):
             return 0
         else:
-            return self.statistics['link_fre'][existing_features["true_link"]]
+            return statistics['link_fre'][existing_features["true_link"]]
 
     def _get_symmetry(self, sys_info: SysOutputInfo, existing_features: dict):
         if existing_features['true_link'] in self._symmetric_relations:
@@ -278,7 +320,13 @@ https://github.com/ExpressAI/DataLab/blob/main/docs/SDK/add_new_datasets_into_sd
     # --- End feature functions
 
     def _get_true_label(self, data_point: dict):
-        return data_point["true_label"]
+        return data_point["true_" + data_point["predict"]]
 
     def _get_predicted_label(self, data_point: dict):
         return data_point["predictions"]
+
+    def _get_rank_data(self, data_point: dict):
+        if "true_rank" in data_point.keys():
+            return data_point["true_rank"]
+        else:
+            return None
