@@ -1,7 +1,7 @@
 from __future__ import annotations
 
 from explainaboard import TaskType
-from explainaboard.metric import BIOF1ScoreConfig, MetricConfig
+from explainaboard.metric import MetricConfig, SeqF1ScoreConfig
 from explainaboard.processors.processor_registry import register_processor
 from explainaboard.processors.sequence_labeling import SeqLabProcessor
 from explainaboard.utils.span_utils import BIOSpanOps
@@ -18,10 +18,11 @@ class NERProcessor(SeqLabProcessor):
         cls, source_language=None, target_language=None
     ) -> list[MetricConfig]:
         return [
-            BIOF1ScoreConfig(
+            SeqF1ScoreConfig(
                 name='F1',
                 source_language=source_language,
                 target_language=target_language,
+                tag_schema='bio',
             )
         ]
 
