@@ -1,7 +1,7 @@
 from __future__ import annotations
 
 from explainaboard import TaskType
-from explainaboard.metric import BMESF1ScoreConfig, MetricConfig
+from explainaboard.metric import MetricConfig, SeqF1ScoreConfig
 from explainaboard.processors.processor_registry import register_processor
 from explainaboard.processors.sequence_labeling import SeqLabProcessor
 from explainaboard.utils.span_utils import BMESSpanOps
@@ -17,7 +17,14 @@ class CWSProcessor(SeqLabProcessor):
     def default_metrics(
         cls, source_language=None, target_language=None
     ) -> list[MetricConfig]:
-        return [BMESF1ScoreConfig(name='F1')]
+        return [
+            SeqF1ScoreConfig(
+                name='F1',
+                source_language=source_language,
+                target_language=target_language,
+                tag_schema='bmes',
+            )
+        ]
 
     @classmethod
     def default_span_ops(cls) -> BMESSpanOps:
