@@ -2,6 +2,8 @@ import json
 
 import requests
 
+from explainaboard.utils.logging import get_logger
+
 if __name__ == "__main__":
     end_point_upload_dataset = "https://datalab.nlpedia.ai/api/normal_dataset/read_stat"
     data_info = {
@@ -13,11 +15,11 @@ if __name__ == "__main__":
     response = requests.post(end_point_upload_dataset, json=data_info)
 
     message = json.loads(response.text.replace("null", ""))["message"]
-    print(message)
+    get_logger('test').info(message)
     """
     (1) success
     (2) dataset does not exist
     (3) the dataset does not include the information of _stat
     """
     return_content = json.loads(response.content)
-    print(return_content['content'])
+    get_logger('test').info(return_content['content'])
