@@ -1,3 +1,4 @@
+import dataclasses
 import os
 import unittest
 
@@ -10,6 +11,7 @@ class TestMachineTranslation(unittest.TestCase):
     artifact_path = os.path.join(test_artifacts_path, "machine_translation")
     tsv_dataset = os.path.join(artifact_path, "dataset.tsv")
     txt_output = os.path.join(artifact_path, "output.txt")
+    json_output_with_features = os.path.join(artifact_path, "output_with_features.json")
 
     def test_load_tsv(self):
         loader = get_custom_dataset_loader(
@@ -75,6 +77,36 @@ class TestMachineTranslation(unittest.TestCase):
         # condgen features are a subset of MT features
         self.assertDictEqual(mt_features, {**mt_features, **condgen_features_1})
 
+    def test_custom_features(self):
+        loader = get_custom_dataset_loader(
+            TaskType.machine_translation,
+            self.tsv_dataset,
+            self.json_output_with_features,
+            Source.local_filesystem,
+            Source.local_filesystem,
+            FileType.tsv,
+            FileType.json,
+        )
+        data = loader.load()
+        self.assertEqual(len(data), 4)
+        self.assertEqual(
+            data[0],
+            {
+                'source': 'Ak sa chcete dostať ešte hlbšie, môžete si všimnúť '
+                + 'trhlinky.',
+                'reference': 'Now just to get really deep in , you can really get to '
+                + 'the cracks .',
+                'id': '0',
+                'hypothesis': 'If you want to get a deeper , you can see the forces .',
+                'num_capital_letters': 1,
+            },
+        )
+
+        processor = get_processor(TaskType.machine_translation.value)
+
+        sys_info = processor.process(dataclasses.asdict(data.metadata), data.samples)
+        self.assertTrue('num_capital_letters' in sys_info.results.fine_grained)
+
 
 if __name__ == '__main__':
     unittest.main()
